/* Authors: Jiangshan Yu, Cas Cremers */
/*
    =======Entity=======

    - log maintainer, who maintains logs for all of his users,
      e.g. Robert, and is assumed to have an authenticated copy of his
      customer's long-term public key.

    - Robert, who has a pair of long term secret/public key, and is the
      recipient of a message m. We assume that Robert has an
      authenticated copy of the log maintainer's long-term public key.

    - Sally, who wants to send a message m to Robert.

    =======Adversary model=======

    Almost all systems (e.g. Windows or Mac OS) and protocols
    (e.g. TLS) suffer from an undesired life cycle of security:

    software released -> vulnerability found and possibly exploited ->
    software updated and exploit eliminated -> new vulnerability found
    and possibly exploited ->...

    Based on the above life cycle of security, we consider an
    adversary who can get all (long-term and short-term) secrets of
    victims. The attacker can repeatedly fully control the victim's
    device, but each time only for a limited period.  For example,
    after a system (e.g. Windows) or a protocol (e.g. OpenSSL) has
    been released, there are always some undiscovered security
    flaws. During the time that the system has not been patched, an
    attacker is able to get some credential (e.g. a secret key) out
    from a victim's device. However, some time later, the security
    flaw is discovered and fixed, and the attacker's ability to
    control the device is removed.

    The security properties we want to guarantee is that if a
    compromised long-term secret will have exposed future messages,
    then the genuine owner of the long-term secret can detect the fact
    of the compromised secret.

    =======Protocol=======

    The basic idea is that the commitments (e.g. signatures) about the
    usage of long-term private keys are recorded in a log, and the
    sender ONLY uses a key if the associated usage is recorded in the
    log. This enables a key owner to monitor the usage of a key. So,
    when a long-term key is compromised and abused by an attacker, the
    genuine key owner can detect that the unauthorised usage is
    recorded in the log.

    - Init Robert:
        Robert is issued a secure device and a pair of long term keys.

    - Init Log maintainer:
        L is issued a long term key. (We only allow one log maintainer
        in this analysis, to reduce the verification space; it can be
        extended to multiple log maintainers.)

    - Init logs for Robert:
        Two empty logs are generated for Robert, one is stored in
        Robert's device, and the other one is maintained by the log
        maintainer in the cloud. We only allow one log for each user
        to reduce the verification space. It can be extended to
        multiple logs for one user, e.g. one log for each long term
        key of a user.

    - Update ephemeral keys:

        1. Robert generates a set of short-term keys, signs them
           together with the root hash of the local log, sends the
           request together with signatures to the log maintainer; (In
           this analysis, we only ask Robert to generate and update one
           key each time rather than a set of keys.)

        2. Log maintainer verifies the signed request, signs the
           confirmation that he has inserted the data into the log,
           and sends the confirmation back;

        3. Robert verifies the confirmation, updates keys and his log
           in the device, and removes past ephemeral secret keys
           from his device.

    - Secure communication:

       Sally gets Robert's certificate of his short-term public key stpk
       from the log, verifies the signature, encrypts message m by
       using stpk, then sends the ciphertext to Robert. Robert can
       decrypt is by using associated ephemeral secret key, and update
       his keys with the log maintainer again. Note that in this
       verification we require each key can be used only once. In
       practice, each ephemeral key would probably be used multiple
       times.

    =======Notes=======

    Rule name: All capital, with '_' to separate words, e.g. RULE_NAME

    Actions: First letter is capital, with '_' to separate words, e.g. Action_Name

    State facts: Eg, 'StSecureCommA1' start with 'St', ends in role name
    and step in role that outputs it. By convention, the first argument
    is the thread identifier (tid), freshly generated in the first step
    of the role.

    Conclusion:  First letter is capital, without '_', e.g. ConclusionName

    Lemma and axiom: lower case letter only, with '_' to separate words, e.g.lemma_name

*/

//theory DECIM
begin

builtins: multiset
functions: adec/2, aenc/2, fst/1, h/1, pair/2, pk/1, sdec/2, senc/2,
  sign/2, snd/1, false/0, true/0, verify/3
equations:
    adec(aenc(x.1, pk(x.2)), x.2) = x.1,
    fst(<x.1, x.2>) = x.1,
    sdec(senc(x.1, x.2), x.2) = x.1,
    snd(<x.1, x.2>) = x.2,
    verify(sign(x.1, x.2), x.1, pk(x.2)) = true

/*Initialisation: key owner. */

rule INIT_R:
     [ Fr(~ltkR),
       Fr(~devid) ]
   --[ Is_Type('KeyOwner', $R),
       R_Key($R, pk(~ltkR), ~ltkR),
       LogDeviceStatus(~devid, $R, 'nil'),
       Start_Role('R',~devid)
     ]->
     [ !Ltk($R, ~ltkR),
       !Pk($R, pk(~ltkR)),
       Out(pk(~ltkR)),
       Device(~devid, $R, ~ltkR, 'secure'), // a user's device is initially
                                   // secure, i.e. no vulnerability
                                   // found so far.
       LogDevice(~devid, $R, 'nil')  // the log stored in Robert's device
     ]

/*Initialisation: log server. */

rule INIT_L:
    [ Fr(~ltkL) ]
  --[ Is_Type('LogMaintainer', $L),
      Only_One('INITL') // we only allow to have one log maintainer $L
    ]->
    [ !Ltk($L, ~ltkL),
      !Pk($L, pk(~ltkL)),
      Out(pk(~ltkL))
    ]


rule INIT_LOG:
     [ 
       Fr(~lcid)
     ]
   --[ Is_Type('LogMaintainer', $L),
       Is_Type('KeyOwner', $R),
       Only_One('INITLOG'), // each user can only have one log with the
                          // log maintainer; which is a reasonable
                          // restriction.
       LogCloudStatus(~lcid, $R, 'nil','start')
     ]->
     [
      LogCloud(~lcid, $R, 'nil') // the log maintained by the log maintainer
                          // in the cloud
     ]

/* Robert generates short-term keys, signs them, sends request to the
   log maintainer.*/

rule UPDATE_EPHEMERAL_KEY_R_1:
  let cert_R = sign( <'cert', pk(~stk), $R>, ltkR)
      Req = sign( <'UpdateRequest', cert_R, h(LogD) >, ltkR)
  in
    [ Fr(~stk), //stk is the short term (ephemeral) key
      !Ltk($R, ltkR),
      LogDevice(devid, $R, LogD)
    ]
  --[  Is_Type('LogMaintainer', $L),
       Is_Type('KeyOwner', $R),
       DeviceSecret(devid, $R, ~stk),
       Start_Role(<'UE','B'>,<$R,$L>)
    ]->
    [ LogToAdd($R, cert_R),
      LogDevice(devid, $R, LogD),
      StUpdateEphKeyR1(~stk, $R, ltkR, $L, ~stk, devid),
      Out(Req),
      Out(<pk(~stk),h(LogD)>)	// [USS] Unfold potential secrets in signature
    ]


  /*  Log maintainer verifies the signed request, signs the
    confirmation that he has inserted the data into the log; */

rule UPDATE_EPHEMERAL_KEY_L_1:
  let
      cert_R = sign( <'cert', stpk, $R> , ltkR)
      LogCNew = LogC+<$R, cert_R>
      Req = sign( <'UpdateRequest', cert_R, h(LogD) >, ltkR)
      Confirmation = sign( (< 'Confirmation', $R, $L, h(LogCNew), cert_R> ), ltkL)
  in
    [ In(Req),
      !Ltk($L, ltkL),
      !Pk($R, ltpkR),
      LogCloud(lcid, $R, LogC)
    ]
  --[ Eq(verify(cert_R, <'cert',stpk, $R>, ltpkR), true ),
      Eq(verify(Req, <'UpdateRequest', cert_R, h(LogD)>, ltpkR), true ),
      //Eq(h(LogC), h(LogD)), // to avoid replay attacks
      Is_Type('LogMaintainer', $L),
      Is_Type('KeyOwner', $R),
      LogCloudStatus(lcid, $R, LogCNew, 'update'),
      Start_Role(<'UE','L'>,<$R,$L>)
    ]->
    [
      Out(Confirmation),
      CurrentCert($R, cert_R), // this fact shows the current valid
                               // certificate of Robert
      LogCloud(lcid, $R, LogCNew),
      Out(h(LogCNew))		// [USS] Unfold potential new secrets in signature
    ]

/* Robert verifies the confirmation, and updates keys and log in the
   device; */

rule UPDATE_EPHEMERAL_KEY_R_2:
  let
      cert_R = sign(<'cert',stpk, $R>, ltkR)
      Req = sign( <'UpdateRequest', cert_R, h(LogD)>, ltkR)
      Confirmation = sign( <'Confirmation', $R, $L, h(LogC), cert_R>, ltkL)
      LogDNew = LogD+<$R, cert_R>
  in
    [ StUpdateEphKeyR1(tid, $R, ltkR, $L, stk, devid),
      In(Confirmation),
      !Pk($L, ltpkL),
      LogToAdd($R, cert_R),
      LogDevice(devid, $R, LogD)
    ]
  --[ Is_Type('KeyOwner', $R),
      Is_Type('LogMaintainer', $L),
      Eq(verify(Confirmation, <'Confirmation', $R, $L, h(LogC), cert_R>, ltpkL), true),
      Eq(h(LogC), h(LogD+<$R, cert_R>)),
      Eq(stpk,pk(stk)),
      LogDeviceStatus(devid, $R, LogDNew)
    ]->
    [
     StoresKey(devid, $R, stk, cert_R), //this fact represents that Robert
                                 //stores the short-term key in his
                                 //device, the associated certificate
                                 //is cert_R
     LogDevice(devid, $R, LogDNew) // append the new cert into
                                        // his log
    ]

/* An attacker is allowed to compromised a device, and then get the
   long term secret and short term secret stored in it. 
   We also add two individual rules to simplify some of the property
   specifications.
 */

rule COMPROMISE_DEVICE_BOTH:
   let
     cert_R = sign(<'cert',pk(stk), R>, ltkR)
   in
    [ !Ltk(R, ltkR),
      StoresKey(devid, R, stk, cert_R),
      Device(devid, R, ltkR, 'secure')
    ]
  --[ Compromise_Device(devid, R, ltkR, stk),
      Is_Type('KeyOwner', R),
      Start_Role('Compromise',<R,devid>)
    ]->
    [ Out(<stk,ltkR>),
      StoresKey(devid, R, stk, cert_R),
      Device(devid, R, ltkR, 'compromised')
    ]

rule COMPROMISE_DEVICE_LTKEY:
    [ !Ltk(R, ltkR),
      Device(devid, R, ltkR, 'secure')
    ]
  --[ Compromise_Device(devid, R, ltkR, 'none'),
      Is_Type('KeyOwner', R),
      Start_Role('Compromise',<R,devid>)
    ]->
    [ Out(ltkR),
      Device(devid, R, ltkR, 'compromised')
    ]

/* Device can be patched. To reduce the verification space and the
   number of open chains, we only allow to fix the device once. In
   other words, rather than having a repected cycle

   Secure--compromised--fixed--compromised again -- fixed...

   we have only one part of the cycle, namely
   secure--compromised--fixed--compromised
   and the device will remain to be broken. In theory, the
   verification should remain the same.*/

rule FIX_DEVICE:
       [ !Ltk(R, ltkR),
         Device(devid, R, ltkR, 'compromised')
       ]
     --[ Is_Type('KeyOwner', R),
         Only_One('FIXDEVICE')
       ]->
       [ Device(devid, R, ltkR, 'secure') ]

/* Sally gets a short-term public key stpk of Robert from the log,
   verifies the signature, encrypts message m by using stpk, then
   sends the ciphertext to Alic. */

rule SECURE_COMMUNICATION_S_1: // Sally requests cert
  let
  m_1=<'CertReq', $R, ~N>
  in
  [ Fr(~N),
    !Pk($R, ltpkR),
    !Pk($L, ltpkL)
  ]
--[ Is_Type('KeyOwner', $R),
    Is_Type('LogMaintainer', $L),
    Start_Role(<'SC','S'>,<$R,$L>)
  ]->
  [ Out(m_1),
    StSecureCommS1(~N, $L, $R, ~N, ltpkR, ltpkL)
  ]

rule SECURE_COMMUNICATION_L_1: // Log maintainer finds the current valid
                             // cert, and give it to Sally.
  let
  m_1=<'CertReq', $R, N>
  cert_R = sign(<'cert',stpk, $R>, ltkR)
  m_2=sign(<'CertResp',$R, cert_R, N>, ltkL)
  in
  [ In(m_1),
    CurrentCert($R, cert_R),
    !Ltk($L, ltkL)
  ]
--[ Is_Type('KeyOwner', $R),
    Is_Type('LogMaintainer', $L),
    Start_Role(<'SC','L'>,<$R,$L>)
  ]->
  [ Out(m_2)
  ]

// rule COMPROMISE_LOG:
//   [ !Ltk($L, ltkL)
//   ]
// --[ Is_Type('LogMaintainer', $L),
//     Compromise_Log($L,ltkL)
//   ]->
//   [ Out(ltkL)
//   ]
  
rule SECURE_COMMUNICATION_S_2: // Sally verifies the cert, and sends m
                             // encrypted with the associated stpk
  let
  cert_R = sign(<'cert',stpk, $R>, ltkR)
  m_2 = sign(<'CertResp',$R, cert_R, N>, ltkL)
  m_3 = aenc{~m}stpk
  in
  [ Fr(~m),
    StSecureCommS1(tid, $L, $R, N, ltpkR, ltpkL),
    In(m_2)
  ]
--[ Eq(verify(m_2, <'CertResp',$R, cert_R, N>, ltpkL), true ),
    Eq(verify(cert_R, <'cert',stpk, $R>, ltpkR), true ),
    Is_Type('KeyOwner', $R),
    Is_Type('LogMaintainer', $L),
    MsgSent($R, ltkR, stpk, ~m)
  ]->
  [
   Out(m_3)
  ]

rule SECURE_COMMUNICATION_B_1: // Robert uses the current valid short
                             // term key to decrypt the received
                             // message.
  let
  cert_R = sign(<'cert',stpk, $R>, ltkR)
  m_3 = aenc{m}stpk
  in
    [ !Ltk($R, ltkR),
      StoresKey(devid, $R, stk, cert_R),
      In(m_3)
    ]
  --[ Is_Type('KeyOwner', $R),
      MsgReceived(devid, $R, ltkR, stk, m),
      Start_Role(<'SC','B'>,<$R,$L>)
    ]->
    [ ]


rule PERIODICAL_VERIFICATION_GOOD:
  [ 
    LogCloud(lcid, $R, LogC),
    LogDevice(devid, $R, LogD)
  ]
--[ CheckedLog(devid, $R, 'good', 'null'),
    SubsetEq(LogC, LogD)		// LogC is a subset of LogD
  ]->	
  [ 
    LogCloud(lcid, $R, LogC),
    LogDevice(devid, $R, LogD)
  ]

rule PERIODICAL_VERIFICATION_BAD:
  let
  cert_R = sign( <'cert',stpk, $R> , ltkR)
  Cuckoo = <$R, cert_R>
  LogC = Cuckoo+rest
  in
  [ LogCloud(lcid, $R, LogC),
    LogDevice(devid, $R, LogD)
  ]
--[ NotIn(Cuckoo, LogD), 
    CheckedLog(devid, $R, 'bad', stpk)   // bad key is in LogC but not in LogD
  ]->
  [ 
    LogCloud(lcid, $R, LogC),
    LogDevice(devid, $R, LogD)
  ]

axiom only_one:
  "(All x #i #j. (Only_One(x)@i & Only_One(x)@j) ==> (#i = #j))"

axiom types_distinct:
  "(All t1 t2 x #i #j. (Is_Type(t1,x)@i & Is_Type(t2,x)@j) ==> (t1 = t2))"

axiom eq_check_succeed:
  "All x y #i. Eq(x,y) @ i ==> x = y"

axiom neq_check_succeed:
  "All x y #i. Neq(x,y) @ i ==> not (x = y)"

axiom notin_check_succeed:
  "All x l #i. NotIn(x,l) @ i ==> not (Ex rest. x+rest = l)"

axiom subseteq_check_succeed:
  "All l m #i. SubsetEq(l,m) @ i ==> ( ( l = m ) | (Ex rest. l+rest = m) )"

// vi:ft=spthy 
